// DISCLAIMER: I am not a C++/Arduino programmer so this code is bad AKA not good :)
// Also I mainly make music go stream my tunes plz :)
// - Nick Newhouse (https://linktr.ee/nicknewhouse)

#include <Adafruit_NeoPixel.h> // library for the individually addressable LEDs I used
#include <EEPROM.h> // used for persisting game state when turned off

#define LEDS_PIN 1
#define SHUFFLE_PIN 7 // randomizer button/switch is on this pin
#define DICE_ROLL_PIN 9 // optional button for playing without dice! long-press to change LED brightness

#define NUM_RESOURCE_LEDS 38 // should be a multiple of 19
#define NUM_PORT_LEDS 9
#define NUM_LEDS (NUM_RESOURCE_LEDS + NUM_PORT_LEDS)
#define MAX_BRIGHTNESS 255
#define INCREMENT_SIZE 3 // controls how quickly the 3:1 ports will strobe

#define REFRESH_DELAY 40 // time (in milliseconds) to pause between re-rendering the LEDs

// Set up the NeoPixel library
Adafruit_NeoPixel pixels(NUM_LEDS, LEDS_PIN, NEO_GRB + NEO_KHZ800);

const int resourceTilesInit [19] = {0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5}; // each integer represents a resource (see below)
const int portsInit [9] = {0, 1, 2, 3, 4, 6, 6, 6, 6};
const int tokensInit [19] = {5, 2, 6, 3, 8, 10, 9, 12, 11, 4, 8, 10, 9, 4, 5, 6, 3, 11, 7};
//const int tokenOrdering [19] = {0, 11, 12, 17, 10, 9, 16, 8, 7, 15, 6, 5, 14, 4, 3, 13, 18, 1, 2}; // my prototype's ordering
const int tokenOrdering [19] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18}; // normal spiral ordering

// These RGB values could technically be anything
// but I chose red, yellow, teal, pink, green, and white (roughly)
const int brick [3] = { 255, 0, 0 }; // 0
const int wheat [3] = { 255, 255, 0 }; // 1
const int ore [3] = { 0, 255, 255 }; // 2
const int sheep [3] = { 255, 31, 255 }; // 3 -- i know it's wool, i like saying sheep though
const int wood [3] = { 0, 255, 0 }; // 4
const int desert [3] = { 255, 255, 255 }; // 5
// 3:1 ports will be represented by 6 and will be a (slow) rainbow strobe

// these arrays will be shuffled from the initialized arrays above
int resourceTiles [19];
int ports [9];

int tokens [19];

float brightness = 1.0; // controls dimming the board
float brightnessIncrement = -0.00001; // controls how quickly the board dims

int rolledTiles [2] = {-1, -1}; // stores indices for resources that the program has rolled
float glowBrightness = 1.0;
float glowIncrement = -0.025; // controls how quickly the rolled numbers blink

// these RGB values control the 3:1 port rainbow strobing
int r = MAX_BRIGHTNESS;
int g = 0;
int b = 0;

// button for shuffling the board
bool shuffleButton = 1;

// button for rolling dice & dimming board
bool diceButton = 1;
unsigned long timePressed = 0;

unsigned long previousRender = 0;

void shuffleBoard() {
  randomSeed(millis() % 1000); // more random than a floating analog pin
  shuffleResourceTiles();
  shufflePorts();
  placeDesertToken();
  resetGlow();
}

void resetGlow() {
  glowIncrement = 0.025;
  glowBrightness = 1.0;
  rolledTiles[0] = -1;
  rolledTiles[1] = -1;
}

void rollDice() { 
  resetGlow();
  int roll = random(1, 7) + random(1, 7);

  int j = 0;
  
  for (int i = 0; i < 19; i++) {
      if (tokens[i] == roll) {
        rolledTiles[j] = i;
        j += 1;
      }
  }

  diceAnimation();
}

void diceAnimation() {
  int ledsPerResourceTile = NUM_RESOURCE_LEDS / 19;
   
  for(int i=0; i+1 < NUM_RESOURCE_LEDS; i+=ledsPerResourceTile) {
     int colorIndex = i / ledsPerResourceTile;
     int color = resourceTiles[tokenOrdering[colorIndex]];

     if (!isGlowing(colorIndex)) {
       for (int j = 0; j < ledsPerResourceTile; j++) {
         pixels.setPixelColor(i+j, 0, 0, 0);
       }
     }
      
     pixels.show();
     delay(20);
  }

  delay(2000);

  float increment = brightness / 50.0;
  float tileBrightness = 0.0;
  
  while (tileBrightness < brightness) {
    for(int i=0; i < NUM_RESOURCE_LEDS; i++) {
      int colorIndex = i / (NUM_RESOURCE_LEDS / 19);
      int color = resourceTiles[tokenOrdering[colorIndex]];

      if (isGlowing(tokenOrdering[colorIndex])) {
        animateLed(i, color, brightness);
      } else {
        animateLed(i, color, tileBrightness);
      }
    }

    pixels.show();
    
    tileBrightness += increment;
    delay(25);
  }
}

void shuffleResourceTiles() {
  for (int i = 0; i < 19; i++) {
     int j = random(0, 19);
     int temp = resourceTiles[j];
     resourceTiles[j] =  resourceTiles[i];
     resourceTiles[i] = temp;
  }
}

void shufflePorts() {    
  for (int i = 0; i < 9; i++) {
     int j = random(0, 9);
     int temp = ports[j];
     ports[j] =  ports[i];
     ports[i] = temp;
  }
}


void placeDesertToken() {
  for (int i = 0; i < 19; i++) {
    tokens[i] = tokensInit[i];
  }

  int desertIndex = -1;
  
  for (int i = 0; i < 19; i++) {
    if (resourceTiles[i] == 5) {
      desertIndex = i;
    }
  }

  for (int i = 18; i >= desertIndex; i--) {
    if (i < 18) {
      tokens[i + 1] = tokens[i];
    }
    
    if (i == desertIndex) {
      tokens[i] = 7;
    }
  }
}

// persist the game state after powering down by using EEPROM
void storeGame() {
  int addressIndex = 0;
  
  for (int i = 0; i < 19; i++) {
    EEPROM.write(addressIndex, resourceTiles[i] >> 8);
    EEPROM.write(addressIndex + 1, resourceTiles[i] & 0xFF);
    addressIndex += 2;
  }

  for (int i = 0; i < 9; i++) {
    EEPROM.write(addressIndex, ports[i] >> 8);
    EEPROM.write(addressIndex + 1, ports[i] & 0xFF);
    addressIndex += 2;
  }
}

void readGame() {
  int addressIndex = 0;
  
  for (int i = 0; i < 19; i++) {
    resourceTiles[i] = (EEPROM.read(addressIndex) << 8) + EEPROM.read(addressIndex + 1);
    addressIndex += 2;
  }

  for (int i = 0; i < 9; i++) {
    ports[i] = (EEPROM.read(addressIndex) << 8) + EEPROM.read(addressIndex + 1);
    addressIndex += 2;
  }

  placeDesertToken();
}

// This logic roughly steps through the color spectrum using RGB values
void nextRainbowColor() {
  if (g < MAX_BRIGHTNESS && r > 0 && b == 0) {
    r -= INCREMENT_SIZE;
    g += INCREMENT_SIZE;
  } else if (b < MAX_BRIGHTNESS && g > 0 && r == 0) {
    g -= INCREMENT_SIZE;
    b += INCREMENT_SIZE;
  } else if (r < MAX_BRIGHTNESS && b > 0 && g == 0) {
    b -= INCREMENT_SIZE;
    r += INCREMENT_SIZE;
  }
}

void updateGlowBrightness() {
  if (glowBrightness >= 1 || glowBrightness <= 0) {
    glowIncrement *= -1;
  }

  float newBrightness = glowBrightness + glowIncrement;

  if (newBrightness < 0) {
    newBrightness = 0.0;
  } else if (newBrightness > 1) {
    newBrightness = 1.0;
  }
 
  glowBrightness = newBrightness;
}

bool isGlowing(int index) {
  return index == rolledTiles[0] || index == rolledTiles[1];
}

void handleButtonInputs() {
  handleShuffleButton();
  handleDiceButton();
}

void handleShuffleButton() {
  bool newShuffleButton = digitalRead(SHUFFLE_PIN);
    
  if (newShuffleButton != shuffleButton) {
    shuffleButton = newShuffleButton;

    if (shuffleButton == 1){
      shuffleBoard();
      storeGame();
    }
  }
}

void handleDiceButton() {
  bool newDiceButton = digitalRead(DICE_ROLL_PIN);
  bool longPress = millis() - timePressed > 1500;

  if (newDiceButton == 0) {
    diceButton = newDiceButton;

    if (timePressed == 0) {
      timePressed = millis();
    } else {
      if (longPress && timePressed != 0) {
        if ((brightness + brightnessIncrement) >= 0.20 && (brightness + brightnessIncrement) <= 1.0) {
          brightness += brightnessIncrement;
        }
      }
    }
  } else {
    if (longPress) {
      if (timePressed != 0) {
        brightnessIncrement = (brightness < 0.60) ? abs(brightnessIncrement) : -abs(brightnessIncrement);
        timePressed = 0;
      }
    } else {
      if (diceButton == 0) {
        rollDice();
        diceButton = newDiceButton;
      }
    }
  }
}

void animateLeds() {
  updateGlowBrightness();

  nextRainbowColor();
//        nextResourceColor();    // the alternative 3:1 port coloring function

  for(int i=0; i < NUM_LEDS; i++) {
    int colorIndex = i / (NUM_RESOURCE_LEDS / 19);
    int color = resourceTiles[tokenOrdering[colorIndex]];

    float tileBrightness = brightness;

    // use the port colors once we've assigned all of the resource tiles 
    if (i >= NUM_RESOURCE_LEDS) {
      colorIndex = i - NUM_RESOURCE_LEDS;
      color = ports[colorIndex];
    } else {
      if (isGlowing(tokenOrdering[colorIndex])) {
        tileBrightness *= glowBrightness;
      }
    }

    animateLed(i, color, tileBrightness);
  }

  pixels.show();   // Send the updated pixel colors to the hardware.
}

void animateLed(int i, int color, float tileBrightness) {
  // here is where the numerical associations with the resource colors are used
  switch(color) {
    case 0:
      pixels.setPixelColor(i, pixels.Color(floor(tileBrightness * brick[0]), floor(tileBrightness * brick[1]), floor(tileBrightness * brick[2])));
      break;
    case 1:
      pixels.setPixelColor(i, pixels.Color(floor(tileBrightness * wheat[0]), floor(tileBrightness * wheat[1]), floor(tileBrightness * wheat[2])));
      break;
    case 2:
      pixels.setPixelColor(i, pixels.Color(floor(tileBrightness * ore[0]), floor(tileBrightness * ore[1]), floor(tileBrightness * ore[2])));
      break;
    case 3:
      pixels.setPixelColor(i, pixels.Color(floor(tileBrightness * sheep[0]), floor(tileBrightness * sheep[1]), floor(tileBrightness * sheep[2])));
      break;
    case 4:
      pixels.setPixelColor(i, pixels.Color(floor(tileBrightness * wood[0]), floor(tileBrightness * wood[1]), floor(tileBrightness * wood[2])));
      break;
    case 5:
      pixels.setPixelColor(i, pixels.Color(floor(tileBrightness * desert[0]), floor(tileBrightness * desert[1]), floor(tileBrightness * desert[2])));
      break;
    case 6:
      pixels.setPixelColor(i, pixels.Color(floor(tileBrightness * r), floor(tileBrightness * g), floor(tileBrightness * b)));
      break;
  }
}

void setup() {
//  Serial.begin(9600);
  previousRender = millis();

  pinMode(SHUFFLE_PIN, INPUT_PULLUP);
  pinMode(DICE_ROLL_PIN, INPUT_PULLUP);

  readGame();
  pixels.begin(); // INITIALIZE NeoPixel strip object (REQUIRED)
}

void loop() {  
  if (millis() - previousRender > REFRESH_DELAY) {
    previousRender = millis();
    animateLeds();
  }
    
  handleButtonInputs();
}
